"""
Functions for optimising the measurement cost of obtaining the expectation value
"""

import networkx as nx
import numpy as np
from qibo import gates


def check_terms_commutativity(term1: str, term2: str, qubitwise: bool):
    """
    Check if terms 1 and 2 are mutually commuting. The 'qubitwise' flag determines if the check is for general
    commutativity (False), or the stricter qubitwise commutativity.

    Args:
        term1/term2: Strings representing a single Pauli term. E.g. "X0 Z1 Y3". Obtained from a Qibo SymbolicTerm as
        ``" ".join(factor.name for factor in term.factors)``.
        qubitwise (bool): Determines if the check is for general commutativity, or the stricter qubitwise commutativity

    Returns:
        bool: Do terms 1 and 2 commute?
    """
    # Get a list of common qubits for each term
    common_qubits = {_term[1:] for _term in term1.split() if _term[0] != "I"} & {
        _term[1:] for _term in term2.split() if _term[0] != "I"
    }
    if not common_qubits:
        return True
    # Get the single Pauli operators for the common qubits for both Pauli terms
    term1_ops = [_op for _op in term1.split() if _op[1:] in common_qubits]
    term2_ops = [_op for _op in term2.split() if _op[1:] in common_qubits]
    if qubitwise:
        # Qubitwise: Compare the Pauli terms at the common qubits. Any difference => False
        return all(_op1 == _op2 for _op1, _op2 in zip(term1_ops, term2_ops))
    # General commutativity:
    # Get the number of single Pauli operators that do NOT commute
    n_noncommuting_ops = sum(_op1 != _op2 for _op1, _op2 in zip(term1_ops, term2_ops))
    # term1 and term2 have general commutativity iff n_noncommuting_ops is even
    return n_noncommuting_ops % 2 == 0


def group_commuting_terms(terms_list, qubitwise):
    """
    Groups the terms in terms_list into as few groups as possible, where all the terms in each group commute
    mutually == Finding the minimum clique cover (i.e. as few cliques as possible) for the graph whereby each node
    is a Pauli string, and an edge exists between two nodes iff they commute.

    This is equivalent to the graph colouring problem of the complement graph (i.e. edge between nodes if they DO NOT
    commute), which this function follows.

    Args:
        terms_list: List of strings. The strings should follow the output from
            ``" ".join(factor.name for factor in term.factors)``, where term is a Qibo SymbolicTerm. E.g. "X0 Z1".
        qubitwise: Determines if the check is for general commutativity, or the stricter qubitwise commutativity

    Returns:
        list: Containing groups (lists) of Pauli strings that all commute mutually
    """
    G = nx.Graph()
    # Complement graph: Add all the terms as nodes first, then add edges between nodes if they DO NOT commute
    G.add_nodes_from(terms_list)
    G.add_edges_from(
        (term1, term2)
        for _i1, term1 in enumerate(terms_list)
        for _i2, term2 in enumerate(terms_list)
        if _i2 > _i1 and not check_terms_commutativity(term1, term2, qubitwise)
    )
    # Solve using Greedy Colouring on NetworkX
    sorted_groups = nx.coloring.greedy_color(G)
    group_ids = set(sorted_groups.values())
    # Sort results so that test results will be replicable
    term_groups = sorted(
        sorted(group for group, group_id in sorted_groups.items() if group_id == _id) for _id in group_ids
    )
    return term_groups


def qwc_measurement_gates(grouped_terms):
    """
    Get the list of (basis rotation) measurement gates to be added to the circuit. The measurements from the resultant
    circuit can then be used to obtain the expectation values of ALL the terms in grouped_terms directly.

    Args:
        grouped_terms (list): List of SymbolicTerms that mutually commutes (qubitwise)

    Returns:
        list: Measurement gates to be appended to the Qibo circuit
    """
    m_gates = {}
    for term in grouped_terms:
        for factor in term.factors:
            if m_gates.get(factor.target_qubit) is None and factor.name[0] != "I":
                m_gates[factor.target_qubit] = gates.M(factor.target_qubit, basis=type(factor.gate))
    return list(m_gates.values())


def qwc_measurements(terms_list):
    """
    Sort out a list of Hamiltonian terms into separate groups of mutually qubitwise commuting terms, and returns the
    grouped terms along with their associated measurement gates

    Args:
        terms_list: Iterable of SymbolicTerms

    Returns:
        list: List of two-tuples, with each tuple given as ([`list of measurement gates`], [term1, term2, ...]), where
            term1, term2, ... are SymbolicTerms.
    """
    ham_terms = {" ".join(factor.name for factor in term.factors): term for term in terms_list}
    term_groups = group_commuting_terms(ham_terms.keys(), qubitwise=True)
    result = [
        (qwc_measurement_gates(symbolic_terms := [ham_terms[term] for term in term_group]), symbolic_terms)
        for term_group in term_groups
    ]
    return result


def measurement_basis_rotations(hamiltonian, grouping=None):
    """
    Split up and sort the Hamiltonian terms to get the basis rotation gates to be applied to a quantum circuit for the
    respective (group of) terms in the Hamiltonian

    Args:
        hamiltonian (SymbolicHamiltonian): Hamiltonian of interest
        grouping: Whether or not to group the X/Y terms together, i.e. use the same set of measurements to get the
            expectation values of a group of terms simultaneously. Default value of ``None`` will not group any terms
            together, while ``"qwc"`` will group qubitwise commuting terms together, and return the measurement gates
            associated with each group of X/Y terms

    Returns:
        list: List of two-tuples, with each tuple given as ([`list of measurement gates`], [term1, term2, ...]), where
            term1, term2, ... are SymbolicTerms. The first tuple always corresponds to all the Z terms present, which
            will be two empty lists - ``([], [])`` - if there are no Z terms present.
    """
    result = []
    # Split up the Z and X/Y terms
    z_only_terms = [
        term for term in hamiltonian.terms if not any(factor.name[0] in ("X", "Y") for factor in term.factors)
    ]
    xy_terms = [term for term in hamiltonian.terms if term not in z_only_terms]
    # Add the Z terms into result first, followed by the terms with X/Y's
    if z_only_terms:
        result.append((qwc_measurement_gates(z_only_terms), z_only_terms))
    else:
        result.append(([], []))
    if xy_terms:
        if grouping is None:
            result += [(qwc_measurement_gates([term]), [term]) for term in xy_terms]
        elif grouping == "qwc":
            result += qwc_measurements(xy_terms)
        else:
            raise NotImplementedError("Not ready yet!")
    return result


def allocate_shots(grouped_terms, n_shots, method=None, max_shots_per_term=None):
    """
    Allocate shots to each group of terms in the Hamiltonian for calculating the expectation value of the Hamiltonian.

    Args:
        grouped_terms (list): Output of measurement_basis_rotations
        n_shots (int): Total number of shots to be allocated
        method (str): How to allocate the shots. The available options are: ``"c"``/``"coefficients"``: ``n_shots`` is
            distributed based on the relative magnitudes of the term coefficients, ``"u"``/``"uniform"``: ``n_shots``
            is distributed evenly amongst each term. Default value: ``"c"``.
        max_shots_per_term (int): Upper limit for the number of shots allocated to an individual group of terms. If not
            given, will be defined as a fraction (largest coefficient over the sum of all coefficients in the
            Hamiltonian) of ``n_shots``.

    Returns:
        list: A list containing the number of shots to be used for each group of Pauli terms respectively.
    """
    if method is None:
        method = "c"
    if max_shots_per_term is None:
        # Define based on the fraction of the term group with the largest coefficients w.r.t. sum of all coefficients.
        term_coefficients = np.array(
            [sum(abs(term.coefficient.real) for term in terms) for (_, terms) in grouped_terms]
        )
        max_shots_per_term = int(np.ceil(n_shots * (np.max(term_coefficients) / sum(term_coefficients))))
    max_shots_per_term = min(n_shots, max_shots_per_term)  # Don't let max_shots_per_term > n_shots if manually defined

    n_terms = len(grouped_terms)
    shot_allocation = np.zeros(n_terms, dtype=int)

    while True:
        remaining_shots = n_shots - sum(shot_allocation)
        if not remaining_shots:
            break

        # In case n_shots is too high s.t. max_shots_per_term is too low
        # Increase max_shots_per_term, and redo the shot allocation
        if np.min(shot_allocation) == max_shots_per_term:
            max_shots_per_term = min(2 * max_shots_per_term, n_shots)
            shot_allocation = np.zeros(n_terms, dtype=int)
            continue

        if method in ("c", "coefficients"):
            # Split shots based on the relative magnitudes of the coefficients of the (group of) Pauli term(s)
            # and only for terms that haven't reached the upper limit yet
            term_coefficients = np.array(
                [
                    sum(abs(term.coefficient.real) for term in terms) if shots < max_shots_per_term else 0.0
                    for shots, (_, terms) in zip(shot_allocation, grouped_terms)
                ]
            )

            # Normalise term_coefficients, then get an initial distribution of remaining_shots
            term_coefficients /= sum(term_coefficients)
            _shot_allocation = (remaining_shots * term_coefficients).astype(int)
            # Only keep the terms with >0 shots allocated, renormalise term_coefficients, and distribute again
            term_coefficients *= _shot_allocation > 0
            if _shot_allocation.any():
                term_coefficients /= sum(term_coefficients)
                _shot_allocation = (remaining_shots * term_coefficients).astype(int)
            else:
                # For distributing the remaining few shots, i.e. remaining_shots << n_terms
                _shot_allocation = np.array(
                    allocate_shots(grouped_terms, remaining_shots, max_shots_per_term=remaining_shots, method="u")
                )

        elif method in ("u", "uniform"):
            # Uniform distribution of shots for every term. Extra shots are randomly distributed
            _shot_allocation = np.array([remaining_shots // n_terms for _ in range(n_terms)])
            if not _shot_allocation.any():
                _shot_allocation = np.zeros(n_terms)
                _shot_allocation[:remaining_shots] = 1
                np.random.shuffle(_shot_allocation)

        else:
            raise NameError("Unknown method!")

        # Add on to the current allocation, and set upper limit to the number of shots for a given term
        shot_allocation += _shot_allocation.astype(int)
        shot_allocation = np.clip(shot_allocation, 0, max_shots_per_term)

    return shot_allocation.tolist()



from qibochem.ansatz.basis_rotation import basis_rotation_gates
from qibochem.ansatz.hardware_efficient import he_circuit
from qibochem.ansatz.hf_reference import hf_circuit
from qibochem.ansatz.ucc import (
    generate_excitations,
    mp2_amplitude,
    sort_excitations,
    ucc_ansatz,
    ucc_circuit,
)

# TODO: Probably can move some of the functions, e.g. generate_excitations/sort_excitations to a new 'util.py'



"""
Functions for obtaining and transforming the molecular Hamiltonian
"""

from functools import reduce

import openfermion
from qibo import symbols
from qibo.hamiltonians import SymbolicHamiltonian


def fermionic_hamiltonian(oei, tei, constant):
    """
    Build molecular Hamiltonian as an InteractionOperator using the 1-/2- electron integrals

    Args:
        oei: 1-electron integrals in the MO basis
        tei: 2-electron integrals in 2ndQ notation and MO basis
        constant: Nuclear-nuclear repulsion, and inactive Fock energy if HF embedding used

    Returns:
        Molecular Hamiltonian as an InteractionOperator
    """
    oei_so, tei_so = openfermion.ops.representations.get_tensors_from_integrals(oei, tei)
    # tei_so already multiplied by 0.5, no need to include in InteractionOperator
    return openfermion.InteractionOperator(constant, oei_so, tei_so)


def qubit_hamiltonian(fermion_hamiltonian, ferm_qubit_map):
    """
    Converts the molecular Hamiltonian to a QubitOperator

    Args:
        fermion_hamiltonian: Molecular Hamiltonian as a InteractionOperator/FermionOperator
        ferm_qubit_map: Which Fermion->Qubit mapping to use

    Returns:
        qubit_operator : Molecular Hamiltonian as a QubitOperator
    """
    # Map the fermionic molecular Hamiltonian to a QubitHamiltonian
    if ferm_qubit_map == "jw":
        q_hamiltonian = openfermion.jordan_wigner(fermion_hamiltonian)
    elif ferm_qubit_map == "bk":
        q_hamiltonian = openfermion.bravyi_kitaev(fermion_hamiltonian)
    else:
        raise KeyError("Unknown fermion->qubit mapping!")
    q_hamiltonian.compress()  # Remove terms with v. small coefficients
    return q_hamiltonian


def qubit_to_symbolic_hamiltonian(q_hamiltonian):
    """
    Converts a OpenFermion QubitOperator to a Qibo SymbolicHamiltonian

    Args:
        q_hamiltonian: QubitOperator

    Returns:
        qibo.hamiltonians.SymbolicHamiltonian
    """
    symbolic_ham = sum(
        reduce(lambda x, y: x * y, (getattr(symbols, pauli_op)(qubit) for qubit, pauli_op in pauli_string), coeff)
        # Sums over each individual Pauli string in the QubitOperator
        for operator in q_hamiltonian.get_operators()
        # .terms gives one operator as a single-item dictionary, e.g. {((1: "X"), (2: "Y")): 0.33}
        for pauli_string, coeff in operator.terms.items()
    )
    return SymbolicHamiltonian(symbolic_ham)



"""
Driver for obtaining molecular integrals from either PySCF or PSI4
"""

from pathlib import Path

import numpy as np
import openfermion
from qibo.hamiltonians import SymbolicHamiltonian

from qibochem.driver.hamiltonian import (
    fermionic_hamiltonian,
    qubit_hamiltonian,
    qubit_to_symbolic_hamiltonian,
)


class Molecule:
    """
    Class representing a single molecule

    Args:
        geometry (list): Molecular coordinates in OpenFermion format,  e.g. ``[('H', (0.0, 0.0, 0.0)), ('H', (0.0, 0.0, 0.7))]``
        charge (int): Net electronic charge of molecule
        multiplicity (int): Spin multiplicity of molecule, given as 2S + 1, where S is half the number of unpaired electrons
        basis (str): Atomic orbital basis set, used for the PySCF/PSI4 calculations. Default: "STO-3G" (minimal basis)
        xyz_file (str): .xyz file containing the molecular coordinates. The comment line can be used to define the electronic
            charge and spin multiplity if it is given in this format: "{charge} {multiplicity}"
        active: Iterable representing the set of MOs to be included in the quantum simulation
            e.g. ``list(range(3,6))`` for an active space with orbitals 3, 4 and 5.

    """

    def __init__(self, geometry=None, charge=0, multiplicity=1, basis=None, xyz_file=None, active=None):
        # Basic properties
        # Define using the function arguments if xyz_file not given
        if xyz_file is None:
            self.geometry = geometry
            self.charge = charge
            self.multiplicity = multiplicity
        else:
            # Check if xyz_file exists, then fill in the Molecule attributes
            assert Path(f"{xyz_file}").exists(), f"{xyz_file} not found!"
            self._process_xyz_file(xyz_file, charge, multiplicity)
        if basis is None:
            # Default bais is STO-3G
            self.basis = "sto-3g"
        else:
            self.basis = basis

        self.nelec = None  #: Total number of electrons for the molecule
        self.norb = None  #: Number of molecular orbitals considered for the molecule
        self.nso = None  #: Number of molecular spin-orbitals considered for the molecule
        self.e_hf = None  #: Hartree-Fock energy
        self.oei = None  #: One-electron integrals
        self.tei = None  #: Two-electron integrals, order follows the second quantization notation

        self.ca = None
        self.pa = None
        self.da = None
        self.nalpha = None
        self.nbeta = None
        self.e_nuc = None
        self.overlap = None
        self.eps = None
        self.fa = None
        self.hcore = None
        self.ja = None
        self.ka = None
        self.aoeri = None

        # For HF embedding
        self.active = active  #: Iterable of molecular orbitals included in the active space
        self.frozen = None  #: Iterable representing the occupied molecular orbitals removed from the simulation

        self.inactive_energy = None
        self.embed_oei = None
        self.embed_tei = None

        self.n_active_e = None  #: Number of electrons included in the active space if HF embedding is used
        self.n_active_orbs = None  #: Number of spin-orbitals in the active space if HF embedding is used

    def _process_xyz_file(self, xyz_file, charge, multiplicity):
        """
        Reads a .xyz file to obtain and set the molecular coordinates (in OpenFermion format),
            charge, and multiplicity

        Args:
            xyz_file: .xyz file for molecule. Comment line should follow "{charge} {multiplicity}"
        """
        with open(xyz_file, encoding="utf-8") as file_handler:
            # First two lines: # atoms and comment line (charge, multiplicity)
            _n_atoms = int(file_handler.readline())  # Not needed/used

            # Try to read charge and multiplicity from comment line
            split_line = [int(_num) for _num in file_handler.readline().split()]
            if len(split_line) == 2:
                # Format of comment line matches (charge, multiplicity):
                _charge, _multiplicity = split_line
            else:
                # Otherwise, use the default (from __init__) values of 0 and 1
                _charge, _multiplicity = charge, multiplicity

            # Start reading xyz coordinates from the 3rd line onwards
            _geometry = []
            for line in file_handler:
                split_line = line.split()
                # OpenFermion format: [('H', (0.0, 0.0, 0.0)), ('H', (0.0, 0.0, 0.7)), ...]
                atom_xyz = [split_line[0], tuple(float(_xyz) for _xyz in split_line[1:4])]
                _geometry.append(tuple(atom_xyz))

        # Set the class attributes
        self.charge = _charge
        self.multiplicity = _multiplicity
        self.geometry = _geometry

    def run_pyscf(self, max_scf_cycles=50):
        """
        Run a Hartree-Fock calculation with PySCF to obtain molecule quantities and
            molecular integrals

        Args:
            max_scf_cycles: Maximum number of SCF cycles in PySCF
        """
        import pyscf  # pylint: disable=C0415

        # Set up and run PySCF calculation
        geom_string = "".join("{} {:.6f} {:.6f} {:.6f} ; ".format(_atom[0], *_atom[1]) for _atom in self.geometry)
        spin = self.multiplicity - 1  # PySCF spin is 2S
        pyscf_mol = pyscf.gto.M(charge=self.charge, spin=spin, atom=geom_string, basis=self.basis, symmetry="C1")

        pyscf_job = pyscf.scf.RHF(pyscf_mol)
        pyscf_job.max_cycle = max_scf_cycles
        pyscf_job.run()
        # print(dir(pyscf_job))

        # Save results from HF calculation
        self.ca = np.asarray(pyscf_job.mo_coeff)  # MO coeffcients
        self.nalpha = pyscf_mol.nelec[0]
        self.nbeta = pyscf_mol.nelec[1]
        self.nelec = sum(pyscf_mol.nelec)
        self.e_hf = pyscf_job.e_tot  # HF energy
        self.e_nuc = pyscf_mol.energy_nuc()
        self.norb = self.ca.shape[1]
        self.nso = 2 * self.norb
        self.overlap = np.asarray(pyscf_mol.intor("int1e_ovlp"))
        self.eps = np.asarray(pyscf_job.mo_energy)
        self.fa = pyscf_job.get_fock()
        self.hcore = pyscf_job.get_hcore()
        self.ja = pyscf_job.get_j()
        self.ka = pyscf_job.get_k()
        self.aoeri = np.asarray(pyscf_mol.intor("int2e"))

        ca_occ = self.ca[:, 0 : self.nalpha]
        self.pa = ca_occ @ ca_occ.T
        self.da = self.ca.T @ self.overlap @ self.pa @ self.overlap @ self.ca

        # 1-electron integrals
        oei = np.asarray(pyscf_mol.intor("int1e_kin")) + np.asarray(pyscf_mol.intor("int1e_nuc"))
        # oei = np.asarray(pyscf_mol.get_hcore())
        oei = np.einsum("ab,bc->ac", oei, self.ca)
        oei = np.einsum("ab,ac->bc", self.ca, oei)
        self.oei = oei

        # Two electron integrals
        # tei = np.asarray(pyscf_mol.intor('int2e'))
        eri = pyscf.ao2mo.kernel(pyscf_mol, self.ca)
        eri4 = pyscf.ao2mo.restore("s1", eri, self.norb)
        tei = np.einsum("pqrs->prsq", eri4)
        self.tei = tei

    # def run_psi4(self, output=None):
    #     """
    #     Run a Hartree-Fock calculation with PSI4 to obtain the molecular quantities and
    #         molecular integrals

    #     Args:
    #         output: Name of PSI4 output file. ``None`` suppresses the output on non-Windows systems,
    #             and uses ``psi4_output.dat`` otherwise
    #     """
    #     import psi4  # pylint: disable=C0415

    #     # PSI4 input string
    #     chgmul_string = f"{self.charge} {self.multiplicity} \n"
    #     geom_string = "\n".join("{} {:.6f} {:.6f} {:.6f}".format(_atom[0], *_atom[1]) for _atom in self.geometry)
    #     opt1_string = "\n\nunits angstrom\nsymmetry c1\n"
    #     mol_string = f"{chgmul_string}{geom_string}{opt1_string}"
    #     # PSI4 calculation options
    #     opts = {"basis": self.basis, "scf_type": "direct", "reference": "rhf", "save_jk": True}
    #     psi4.core.clean()
    #     psi4.set_memory("500 MB")
    #     psi4.set_options(opts)
    #     # Keep the output file of the PSI4 calculation?
    #     if output is None:
    #         # Doesn't work on Windows!
    #         # See: https://psicode.org/psi4manual/master/api/psi4.core.be_quiet.html
    #         try:
    #             psi4.core.be_quiet()
    #         except RuntimeError:
    #             psi4.core.set_output_file("psi4_output.dat", False)
    #     else:
    #         psi4.core.set_output_file(output, False)

    #     # Run HF calculation with PSI4
    #     psi4_mol = psi4.geometry(mol_string)
    #     ehf, wavefn = psi4.energy("hf", return_wfn=True)

    #     # Save 1- and 2-body integrals
    #     ca = wavefn.Ca()  # MO coefficients
    #     self.ca = np.asarray(ca)
    #     # 1- electron integrals
    #     oei = wavefn.H()  # 'Core' (potential + kinetic) integrals
    #     # Convert from AO->MO basis
    #     oei = np.einsum("ab,bc->ac", oei, self.ca)
    #     oei = np.einsum("ab,ac->bc", self.ca, oei)

    #     # 2- electron integrals
    #     mints = psi4.core.MintsHelper(wavefn.basisset())
    #     tei = np.asarray(mints.mo_eri(ca, ca, ca, ca))  # Need original C_a array, not a np.array
    #     tei = np.einsum("pqrs->prsq", tei)

    #     # Fill in the class attributes
    #     self.nelec = sum(wavefn.nalpha(), wavefn.nbeta())
    #     self.nalpha = wavefn.nalpha()
    #     self.nbeta = wavefn.nbeta()
    #     self.e_hf = ehf
    #     self.e_nuc = psi4_mol.nuclear_repulsion_energy()
    #     self.norb = wavefn.nmo()
    #     self.nso = 2 * self.norb
    #     self.oei = oei
    #     self.tei = tei
    #     self.aoeri = np.asarray(mints.ao_eri())
    #     self.overlap = np.asarray(wavefn.S())
    #     self.eps = np.asarray(wavefn.epsilon_a())
    #     self.fa = np.asarray(wavefn.Fa())
    #     self.hcore = np.asarray(wavefn.H())

    #     self.ja = np.asarray(wavefn.jk().J()[0])
    #     self.ka = np.asarray(wavefn.jk().K()[0])

    #     ca_occ = self.ca[:, 0 : self.nalpha]
    #     self.pa = ca_occ @ ca_occ.T
    #     self.da = self.ca.T @ self.overlap @ self.pa @ self.overlap @ self.ca

    # HF embedding functions
    def _inactive_fock_matrix(self, frozen):
        """
        Returns the full inactive Fock matrix

        Args:
            frozen: Iterable representing the occupied orbitals to be removed from the simulation
        """
        # Copy the original OEI as a starting point
        inactive_fock = np.copy(self.oei)

        # Add the two-electron operator, summed over the frozen orbitals
        # Iterate over the active orbitals
        for _p in range(self.norb):
            for _q in range(self.norb):
                # Iterate over the inactive orbitals
                for _orb in frozen:
                    # Add (2J - K) using TEI (in OpenFermion format)
                    inactive_fock[_p][_q] += 2 * self.tei[_orb][_p][_q][_orb] - self.tei[_orb][_p][_orb][_q]
        return inactive_fock

    def _active_space(self, active, frozen):
        """
        Helper function to check the input for active/frozen space and define the default values
        for them where necessary

        Args:
            active: Iterable representing the active-space for quantum simulation
            frozen: Iterable representing the occupied orbitals to be removed from the simulation

        Returns:
            _active, _frozen: Iterables representing the active/frozen space
        """
        n_orbs = self.norb
        n_occ_orbs = self.nalpha

        _active, _frozen = None, None
        if active is None:
            # No arguments given
            if frozen is None:
                # Default active: Full set of orbitals, frozen: empty list
                _active = list(range(n_orbs))
                _frozen = []
            # Only frozen argument given
            else:
                if frozen:
                    # Non-empty frozen space must be occupied orbitals
                    assert max(frozen) + 1 < n_occ_orbs and min(frozen) >= 0, "Frozen orbital must be occupied orbitals"
                _frozen = frozen
                # Default active: All orbitals not in frozen
                _active = [_i for _i in range(n_orbs) if _i not in _frozen]
        # active argument given
        else:
            # Check that active argument is valid
            assert max(active) < n_orbs and min(active) >= 0, "Active space must be between 0 and the number of MOs"
            _active = active
            # frozen argument not given
            if frozen is None:
                # Default frozen: All occupied orbitals not in active
                _frozen = [_i for _i in range(n_occ_orbs) if _i not in _active]
            # active, frozen arguments both given:
            else:
                # Check that active/frozen arguments don't overlap
                assert not (set(active) & set(frozen)), "Active and frozen space cannot overlap"
                if frozen:
                    # Non-empty frozen space must be occupied orbitals
                    assert max(frozen) + 1 < n_occ_orbs and min(frozen) >= 0, "Frozen orbital must be occupied orbitals"
                # All occupied orbitals have to be in active or frozen
                assert all(
                    _occ in set(active + frozen) for _occ in range(n_occ_orbs)
                ), "All occupied orbitals have to be in either the active or frozen space"
                # Hopefully no more problems with the input
                _frozen = frozen
        return _active, _frozen

    def hf_embedding(self, active=None, frozen=None):
        """
        Turns on HF embedding for a given active/frozen space, and fills in the class attributes:
            ``inactive_energy``, ``embed_oei``, and ``embed_tei``.

        Args:
            active: Iterable representing the active-space for quantum simulation
            frozen: Iterable representing the occupied orbitals to be removed from the simulation
        """
        # Default arguments for active and frozen if no arguments given
        if active is None and frozen is None:
            _active, _frozen = self._active_space(self.active, self.frozen)
        else:
            # active/frozen arguments given, process them using _active_space similarly
            _active, _frozen = self._active_space(active, frozen)
        # Update the class attributes with the checked arguments
        self.active = _active
        self.frozen = _frozen

        # Build the inactive Fock matrix first
        inactive_fock = self._inactive_fock_matrix(self.frozen)

        # Calculate the inactive Fock energy
        # Only want frozen part of original OEI and inactive Fock matrix
        _oei = self.oei[np.ix_(self.frozen, self.frozen)]
        _inactive_fock = inactive_fock[np.ix_(self.frozen, self.frozen)]
        self.inactive_energy = np.einsum("ii->", _oei + _inactive_fock)

        # Keep only the active part
        self.embed_oei = inactive_fock[np.ix_(self.active, self.active)]
        self.embed_tei = self.tei[np.ix_(self.active, self.active, self.active, self.active)]

        # Update other class attributes
        self.n_active_orbs = 2 * len(self.active)
        self.n_active_e = self.nelec - 2 * len(self.frozen)

    def hamiltonian(
        self,
        ham_type=None,
        oei=None,
        tei=None,
        constant=None,
        ferm_qubit_map=None,
    ):
        """
        Builds a molecular Hamiltonian using the one-/two- electron integrals. If HF embedding has been applied,
        (i.e. the ``embed_oei``, ``embed_tei``, and ``inactive_energy`` attributes are all not ``None``), the
        corresponding values for the molecular integrals will be used instead.

        Args:
            ham_type: Format of molecular Hamiltonian returned. The available options are:
                ``("f", "ferm")``: OpenFermion ``FermionOperator``,
                ``("q", "qubit")``: OpenFermion ``QubitOperator``, or
                ``("s", "sym")``: Qibo ``SymbolicHamiltonian`` (default)
            oei: 1-electron integrals (in the MO basis). The default value is the ``oei`` class attribute , unless
                the ``embed_oei`` attribute exists and is not ``None``, then ``embed_oei`` is used.
            tei: 2-electron integrals in the second-quantization notation (and MO basis). The default value is the
                ``tei`` class attribute , unless the ``embed_tei`` attribute exists and is not ``None``, then ``embed_tei``
                is used.
            constant: Constant value to be added to the electronic energy. Mainly used for adding the inactive Fock
                energy if HF embedding was applied. Default: 0.0, unless the ``inactive_energy`` class attribute exists
                and is not ``None``, then ``inactive_energy`` is used.
            ferm_qubit_map: Which fermion to qubit transformation to use.
                Must be either ``jw`` (default) or ``bk``

        Returns:
            Molecular Hamiltonian in the format of choice
        """
        # Define default variables
        if ham_type is None:
            ham_type = "sym"
        if oei is None:
            oei = self.oei if self.embed_oei is None else self.embed_oei
        if tei is None:
            tei = self.tei if self.embed_tei is None else self.embed_tei
        if constant is None:
            constant = 0.0 if self.inactive_energy is None else self.inactive_energy
        if ferm_qubit_map is None:
            ferm_qubit_map = "jw"

        constant += self.e_nuc  # Add nuclear repulsion energy

        # Start with an InteractionOperator
        ham = fermionic_hamiltonian(oei, tei, constant)
        if ham_type in ("f", "ferm"):
            # OpenFermion FermionOperator Hamiltonian
            ham = openfermion.transforms.get_fermion_operator(ham)
            ham.compress()
            return ham
        ham = qubit_hamiltonian(ham, ferm_qubit_map)
        if ham_type in ("q", "qubit"):
            # OpenFermion QubitOperator Hamiltonian
            return ham
        if ham_type in ("s", "sym"):
            # Qibo SymbolicHamiltonian
            return qubit_to_symbolic_hamiltonian(ham)
        raise NameError(f"Unknown {ham_type}!")  # Shouldn't ever reach here

    @staticmethod
    def eigenvalues(hamiltonian):
        """
        Finds the lowest 6 exact eigenvalues of the molecular Hamiltonian
            Note: Uses the ``eigenvalues()`` class method for a Qibo ``SymbolicHamiltonian`` object

        Args:
            hamiltonian: Molecular Hamiltonian, given as a ``FermionOperator``, ``QubitOperator``, or
                ``SymbolicHamiltonian`` (not recommended)
        """
        if isinstance(hamiltonian, (openfermion.FermionOperator, openfermion.QubitOperator)):
            from scipy.sparse import linalg  # pylint: disable=C0415

            hamiltonian_matrix = openfermion.get_sparse_operator(hamiltonian)
            # k argument in eigsh will depend on the size of the Hamiltonian
            n_eigenvals = min(6, hamiltonian_matrix.shape[0] - 2)
            # which=SA and return_eigenvalues=False returns the eigenvalues sorted by absolute value
            eigenvalues = linalg.eigsh(hamiltonian_matrix, k=n_eigenvals, which="SA", return_eigenvectors=False)
            # So need to sort again by their (algebraic) value to get the order: smallest->largest
            return sorted(eigenvalues)
        if isinstance(hamiltonian, SymbolicHamiltonian):
            return hamiltonian.eigenvalues()
        raise TypeError("Type of Hamiltonian unknown")



"""
Functions for obtaining the expectation value for some given circuit and Hamiltonian, either from a state
vector simulation, or from sample measurements
"""

from functools import reduce

import qibo
from qibo.hamiltonians import SymbolicHamiltonian
from qibo.symbols import Z

from qibochem.measurement.optimization import (
    allocate_shots,
    measurement_basis_rotations,
)


def expectation(circuit: qibo.models.Circuit, hamiltonian: SymbolicHamiltonian):
    """
    Expectation value using state vector simulations

    Args:
        circuit (qibo.models.Circuit): Quantum circuit ansatz
        hamiltonian (qibo.hamiltonians.SymbolicHamiltonian): Molecular Hamiltonian

    Returns:
        float: Expectation value of the Hamiltonian for the given circuit
    """
    result = circuit(nshots=1)
    state_ket = result.state()
    return hamiltonian.expectation(state_ket)


def symbolic_term_to_symbol(symbolic_term):
    """Convert a single Pauli word in the form of a Qibo SymbolicTerm to a Qibo Symbol"""
    return symbolic_term.coefficient * reduce(lambda x, y: x * y, symbolic_term.factors, 1.0)


def pauli_term_measurement_expectation(pauli_term, frequencies, qubit_map):
    """
    Calculate the expectation value of a single general Pauli string for some measurement frequencies

    Args:
        pauli_term (SymbolicTerm): Single general Pauli term, e.g. X0*Z2
        frequencies: Measurement frequencies, taken from MeasurementOutcomes.frequencies(binary=True)

    Returns:
        float: Expectation value of pauli_term
    """
    # Replace every (non-I) Symbol with Z, then include the term coefficient
    pauli_z = [Z(int(factor.target_qubit)) for factor in pauli_term.factors if factor.name[0] != "I"]
    z_only_ham = SymbolicHamiltonian(pauli_term.coefficient * reduce(lambda x, y: x * y, pauli_z, 1.0))
    # Can now apply expectation_from_samples directly
    return z_only_ham.expectation_from_samples(frequencies, qubit_map=qubit_map)


def expectation_from_samples(
    circuit: qibo.models.Circuit,
    hamiltonian: SymbolicHamiltonian,
    n_shots: int = 1000,
    group_pauli_terms=None,
    n_shots_per_pauli_term: bool = True,
    shot_allocation=None,
) -> float:
    """
    Calculate expectation value of some Hamiltonian using sample measurements from running a Qibo quantum circuit

    Args:
        circuit (qibo.models.Circuit): Quantum circuit ansatz
        hamiltonian (qibo.hamiltonians.SymbolicHamiltonian): Molecular Hamiltonian
        n_shots (int): Number of times the circuit is run. Default: ``1000``
        group_pauli_terms: Whether or not to group Pauli X/Y terms in the Hamiltonian together to reduce the measurement
            cost. Available options: ``None``: (Default) Hamiltonian terms containing X/Y are not grouped together, and
            ``"qwc"``: Terms that commute qubitwise are grouped together
        n_shots_per_pauli_term (bool): Whether or not ``n_shots`` is used for each Pauli term in the Hamiltonian, or for
            *all* the terms in the Hamiltonian. Default: ``True``; ``n_shots`` are used to get the expectation value for each
            term in the Hamiltonian.
        shot_allocation: Iterable containing the number of shots to be allocated to each term in the Hamiltonian respectively if
            n_shots_per_pauli_term is ``False``. Default: ``None``; shots are allocated based on the magnitudes of the coefficients
            of the Hamiltonian terms.

    Returns:
        float: Hamiltonian expectation value
    """
    # Group up Hamiltonian terms to reduce the measurement cost
    grouped_terms = measurement_basis_rotations(hamiltonian, grouping=group_pauli_terms)

    # Check shot_allocation argument if not using n_shots_per_pauli_term
    if not n_shots_per_pauli_term:
        if shot_allocation is None:
            shot_allocation = allocate_shots(grouped_terms, n_shots)
        assert len(shot_allocation) == len(
            grouped_terms
        ), f"shot_allocation list ({len(shot_allocation)}) doesn't match the number of grouped terms ({len(grouped_terms)})"

    total = 0.0
    for _i, (measurement_gates, terms) in enumerate(grouped_terms):
        if measurement_gates and terms:
            _circuit = circuit.copy()
            _circuit.add(measurement_gates)

            # Number of shots used to run the circuit depends on n_shots_per_pauli_term
            result = _circuit(nshots=n_shots if n_shots_per_pauli_term else shot_allocation[_i])

            frequencies = result.frequencies(binary=True)
            qubit_map = sorted(qubit for gate in measurement_gates for qubit in gate.target_qubits)
            if frequencies:  # Needed because might have cases whereby no shots allocated to a group
                # First term is all Z terms, can use expectation_from_samples directly.
                if _i > 0:
                    total += sum(pauli_term_measurement_expectation(term, frequencies, qubit_map) for term in terms)
                # Otherwise, need to use the general pauli_term_measurement_expectation function
                else:
                    z_ham = SymbolicHamiltonian(sum(symbolic_term_to_symbol(term) for term in terms))
                    total += z_ham.expectation_from_samples(frequencies, qubit_map=qubit_map)
    # Add the constant term if present. Note: Energies (in chemistry) are all real values
    total += hamiltonian.constant.real
    return total



"""
Hardware efficient ansatz, e.g. Kandala et al. (https://doi.org/10.1038/nature23879)
"""

from qibo import Circuit, gates


def he_circuit(n_qubits, n_layers, parameter_gates=None, coupling_gates="CZ"):
    """
    Builds the generalized hardware-efficient ansatz, in which the rotation and entangling gates used can be
    chosen by the user

    Args:
        n_qubits: Number of qubits in the quantum circuit
        n_layers: Number of layers of rotation and entangling gates
        parameter_gates: Iterable of single-qubit rotation gates to be used in the ansatz. The gates should be given as
            strings representing valid ``Qibo`` one-qubit gates. Default: ``["RY", "RZ"]``
        coupling_gates: String representing the two-qubit entangling gate to be used in the ansatz; should be a
            valid two-qubit ``Qibo`` gate. Default: ``"CZ"``

    Returns:
        Qibo ``Circuit``: Circuit corresponding to the hardware-efficient ansatz
    """
    # Default variables
    if parameter_gates is None:
        parameter_gates = ["RY", "RZ"]

    circuit = Circuit(n_qubits)

    for _ in range(n_layers):
        # Rotation gates
        circuit.add(getattr(gates, rgate)(qubit, theta=0.0) for qubit in range(n_qubits) for rgate in parameter_gates)

        # Entanglement gates
        circuit.add(getattr(gates, coupling_gates)(qubit, qubit + 1) for qubit in range(n_qubits - 1))
        circuit.add(getattr(gates, coupling_gates)(n_qubits - 1, 0))

    return circuit



"""
Circuit representing an unitary rotation of the molecular (spin-)orbital basis set
"""

import numpy as np
from qibo import gates, models
from scipy.linalg import expm

from qibochem.ansatz import ucc
from qibochem.driver import hamiltonian

# Helper functions


def unitary(occ_orbitals, virt_orbitals, parameters=None):
    r"""
    Returns the unitary rotation matrix :math: `U = \exp(\kappa)` mixing the occupied and virtual orbitals.
    Orbitals are arranged in alternating spins, e.g. for 4 occupied orbitals [0,1,2,3], the spins are arranged as [0a, 0b, 1a, 1b]
    Dimension for the array of rotation parameters is len(occ_orbitals_a)*len(virt_orbitals_a) + len(occ_orbitals_b)*len(virt_orbitals_b)
    The current implementation of this function only accommodates systems with all electrons paired, number of alpha and beta spin electrons are equal
    Args:
        occ_orbitals: Iterable of occupied orbitals
        virt_orbitals: Iterable of virtual orbitals
        parameters: List/array of rotation parameters
            dimension = len(occ_orbitals)*len(virt_orbitals)
    Returns:
        exp(k): Unitary matrix of Givens rotations, obtained by matrix exponential of skew-symmetric
                kappa matrix
    """

    # conserve_spin has to be true for SCF/basis_rotation cases, else expm(k) is not unitary
    ov_pairs = ucc.generate_excitations(1, occ_orbitals, virt_orbitals, conserve_spin=True)
    # print('ov_pairs presort', ov_pairs)
    ov_pairs = ucc.sort_excitations(ov_pairs)
    # print('ov_pairs sorted ', ov_pairs)
    n_theta = len(ov_pairs)
    if parameters is None:
        print("basis rotation: parameters not specified")
        print("basis rotation: using default occ-virt rotation parameter value = 0.0")
        parameters = np.zeros(n_theta)
    elif isinstance(parameters, float):
        print("basis rotation: using uniform value of", parameters, "for each parameter value")
        parameters = np.full(n_theta, parameters)
    else:
        if len(parameters) != n_theta:
            raise IndexError("parameter array specified has bad size or type")
        print("basis rotation: loading parameters from input")

    n_orbitals = len(occ_orbitals) + len(virt_orbitals)
    kappa = np.zeros((n_orbitals, n_orbitals))

    for _i, (_occ, _virt) in enumerate(ov_pairs):
        kappa[_occ, _virt] = parameters[_i]
        kappa[_virt, _occ] = -parameters[_i]

    return expm(kappa), parameters


# clements qr routines
def givens_qr_decompose(U):
    r"""
    Clements scheme QR decompose a unitary matrix U using Givens rotations
    see arxiv:1603.08788
    Args:
        U: unitary rotation matrix
    Returns:
        z_angles: array of rotation angles
        U: final unitary after QR decomposition, should be an identity
    """

    def move_step(U, row, col, row_change, col_change):
        """
        internal function to move a step in Givens QR decomposition of
        unitary rotation matrix
        """
        row += row_change
        col += col_change
        return U, row, col

    def row_op(U, row, col):
        """
        internal function to zero out a row using Givens rotation
        with angle z
        """
        Uc = U.copy()
        srow = row - 1
        scol = col
        z = np.arctan2(-Uc[row][col], Uc[srow][scol])
        temp_srow = Uc[srow, :]
        temp_row = Uc[row, :]
        new_srow = np.cos(z) * temp_srow - np.sin(z) * temp_row
        new_row = np.sin(z) * temp_srow + np.cos(z) * temp_row
        Uc[srow, :] = new_srow
        Uc[row, :] = new_row
        return z, Uc

    def col_op(U, row, col):
        """
        internal function to zero out a column using Givens rotation
        with angle z
        """
        Uc = U.copy()
        srow = row
        scol = col + 1
        z = np.arctan2(-Uc[row][col], Uc[srow][scol])
        temp_scol = Uc[:, scol]
        temp_col = Uc[:, col]
        new_scol = np.cos(z) * temp_scol - np.sin(z) * temp_col
        new_col = np.sin(z) * temp_scol + np.cos(z) * temp_col
        Uc[:, scol] = new_scol
        Uc[:, col] = new_col
        return z, Uc

    N = len(U[0])

    z_array = []

    # start QR from bottom left element
    row = N - 1
    col = 0
    z, U = col_op(U, row, col)
    z_array.append(z)
    # traverse the U matrix in diagonal-zig-zag manner until the main
    # diagonal is reached
    # if move = up, do a row op
    # if move = diagonal-down-right, do a row op
    # if move = right, do a column op
    # if move = diagonal-up-left, do a column op
    while row != 1:
        U, row, col = move_step(U, row, col, -1, 0)
        z, U = row_op(U, row, col)
        z_array.append(z)
        while row < N - 1:
            U, row, col = move_step(U, row, col, 1, 1)
            z, U = row_op(U, row, col)
            z_array.append(z)
        if col != N - 2:
            U, row, col = move_step(U, row, col, 0, 1)
            z, U = col_op(U, row, col)
            z_array.append(z)
        # else:
        #    break
        while col > 0:
            U, row, col = move_step(U, row, col, -1, -1)
            z, U = col_op(U, row, col)
            z_array.append(z)

    return z_array, U


def basis_rotation_layout(N):
    r"""
    generates the layout of the basis rotation circuit for Clements scheme QR decomposition
    Args:
        N: number of qubits/modes
    Returns:
        A:  NxN matrix, with -1 being null, 0 is the control and integers
            1 or greater being the index for angles in clements QR decomposition of
            the unitary matrix representing the unitary transforms that
            rotate the basis
    """

    def move_step(A, row, col, row_change, col_change):
        """
        internal function to move a step in gate layout of br circuit
        """
        row += row_change
        col += col_change
        return A, row, col, updown

    def check_top(A, row, col, row_change, col_change):
        """
        check if we are at the top of layout matrix,
        return false if we are at the top, i.e. row == 1
        (row 0 is not assigned any operation; just control)
        """
        move_step(A, row, col, row_change, col_change)
        return row > 1

    def check_bottom(A, row, col, row_change, col_change):
        """
        check if we are at the bottom of layout matrix
        return false if we are at the bottom, i.e. row == N-1
        """
        N = len(A)
        move_step(A, row, col, row_change, col_change)
        return row < N - 1

    def jump_right(A, row, col, updown):
        """ """
        N = len(A)
        row = N - 2 - col
        col = N - 1
        updown = -1
        return A, row, col, updown

    def jump_left(A, row, col, updown):
        N = len(A)
        row = N + 1 - col
        col = 0
        updown = 1
        return A, row, col, updown

    def assign_element(A, row, col, k):
        A[row - 1][col] = 0
        A[row][col] = k
        return A, row, col

    nangles = (N - 1) * N // 2  # half-triangle
    A = np.full([N, N], -1, dtype=int)

    # first step
    row = 1
    col = 0
    k = 1
    A, row, col = assign_element(A, row, col, k)
    k += 1
    updown = 1

    while k <= nangles:

        if updown == 1:
            if check_top(A, row, col, -1, 1):
                A, row, col, updown = move_step(A, row, col, -1, 1)
                A, row, col = assign_element(A, row, col, k)
            else:  # jump
                # print('jump_right')
                A, row, col, updown = jump_right(A, row, col, updown)
                A, row, col = assign_element(A, row, col, k)
        elif updown == -1:
            if check_bottom(A, row, col, 1, -1):
                A, row, col, updown = move_step(A, row, col, 1, -1)
                A, row, col = assign_element(A, row, col, k)
            else:  # jump
                # print('jump left')
                A, row, col, updown = jump_left(A, row, col, updown)
                A, row, col = assign_element(A, row, col, k)
        # else:
        #    raise ValueError("Bad direction")

        # print(row, col)
        k += 1

    return A


def basis_rotation_gates(A, z_array, parameters):
    r"""
    places the basis rotation gates on circuit in the order of Clements scheme QR decomposition

    Args:
        A:
            NxN matrix, with -1 being null, 0 is the control and integers
            1 or greater being the index for angles in clements QR decomposition of
            the unitary matrix representing the unitary transforms that
            rotate the basis
        z_array:
            array of givens rotation angles in order of traversal from
            QR decomposition
        parameters:
            array of parameters in order of traversal from QR decomposition
    Outputs:
        gate_list:
            list of gates which implement the basis rotation using Clements scheme QR decomposition
        ordered_angles:
            list of angles ordered by sequence of singles excitation gates added to circuit

    """
    N = len(A[0])
    gate_list = []
    ordered_angles = []
    #
    for j in range(N):
        for i in range(N):
            if A[i][j] == 0:
                # gate_list.append(gates.CNOT(i + 1, i))
                # gate_list.append(gates.CRY(i, i + 1, z_array[A[i + 1][j] - 1]))
                gate_list.append(gates.GIVENS(i + 1, i, z_array[A[i + 1][j] - 1]))
                ordered_angles.append(z_array[A[i + 1][j] - 1])
                # gate_list.append(gates.CNOT(i + 1, i))

    return gate_list, ordered_angles



"""
Circuit representing the Unitary Coupled Cluster ansatz in quantum chemistry
"""

import numpy as np
import openfermion
from qibo import Circuit, gates

from qibochem.ansatz.hf_reference import hf_circuit


def expi_pauli(n_qubits, pauli_string, theta):
    """
    Build circuit representing exp(i*theta*pauli_string)

    Args:
        n_qubits: No. of qubits in the quantum circuit
        pauli_string: String in the format: ``"X0 Z1 Y3 X11"``
        theta: Real number

    Returns:
        circuit: Qibo Circuit object representing exp(i*theta*pauli_string)
    """
    # Split pauli_string into the old p_letters format
    pauli_ops = sorted(((int(_op[1:]), _op[0]) for _op in pauli_string.split()), key=lambda x: x[0])
    n_pauli_ops = len(pauli_ops)

    # Convert theta into a real number for applying with a RZ gate
    rz_parameter = -2.0 * theta

    # Generate the list of basis change gates using the pauli_ops list
    basis_changes = []
    for qubit, pauli_op in pauli_ops:
        if pauli_op == "Y":
            basis_changes.append(gates.S(qubit).dagger())
        if pauli_op not in ("I", "Z"):
            basis_changes.append(gates.H(qubit))

    # Build the circuit
    circuit = Circuit(n_qubits)
    # 1. Change to X/Y where necessary
    circuit.add(basis_changes)
    # 2. Add CNOTs to all pairs of qubits in pauli_ops, starting from the last letter
    circuit.add(gates.CNOT(pauli_ops[_i][0], pauli_ops[_i - 1][0]) for _i in range(n_pauli_ops - 1, 0, -1))
    # 3. Add RZ gate to last element of pauli_ops
    circuit.add(gates.RZ(pauli_ops[0][0], rz_parameter))
    # 4. Add CNOTs to all pairs of qubits in pauli_ops
    circuit.add(gates.CNOT(pauli_ops[_i + 1][0], pauli_ops[_i][0]) for _i in range(n_pauli_ops - 1))
    # 3. Change back to the Z basis
    circuit.add(_gate.dagger() for _gate in reversed(basis_changes))
    return circuit


def ucc_circuit(n_qubits, excitation, theta=0.0, trotter_steps=1, ferm_qubit_map=None):
    r"""
    Circuit corresponding to the unitary coupled-cluster ansatz for a single excitation

    Args:
        n_qubits: Number of qubits in the quantum circuit
        excitation: Iterable of orbitals involved in the excitation; must have an even number of elements
            E.g. ``[0, 1, 2, 3]`` represents the excitation of electrons in orbitals ``(0, 1)`` to ``(2, 3)``
        theta: UCC parameter. Defaults to 0.0
        trotter_steps: Number of Trotter steps; i.e. number of times the UCC ansatz is applied with ``theta`` = ``theta / trotter_steps``. Default: 1
        ferm_qubit_map: Fermion-to-qubit transformation. Default is Jordan-Wigner (``jw``).

    Returns:
        Qibo ``Circuit``: Circuit corresponding to a single UCC excitation
    """
    # Check size of orbitals input
    n_orbitals = len(excitation)
    assert n_orbitals % 2 == 0, f"{excitation} must have an even number of items"
    # Reverse sort orbitals to get largest-->smallest
    sorted_orbitals = sorted(excitation, reverse=True)

    # Define default mapping
    if ferm_qubit_map is None:
        ferm_qubit_map = "jw"

    # Define the UCC excitation operator corresponding to the given list of orbitals
    fermion_op_str_template = f"{(n_orbitals//2)*'{}^ '}{(n_orbitals//2)*'{} '}"
    fermion_operator_str = fermion_op_str_template.format(*sorted_orbitals)
    # Build the FermionOperator and make it unitary
    fermion_operator = openfermion.FermionOperator(fermion_operator_str)
    ucc_operator = fermion_operator - openfermion.hermitian_conjugated(fermion_operator)

    # Map the FermionOperator to a QubitOperator
    if ferm_qubit_map == "jw":
        qubit_ucc_operator = openfermion.jordan_wigner(ucc_operator)
    elif ferm_qubit_map == "bk":
        qubit_ucc_operator = openfermion.bravyi_kitaev(ucc_operator)
    else:
        raise KeyError("Fermon-to-qubit mapping must be either 'jw' or 'bk'")

    # Apply the qubit_ucc_operator 'trotter_steps' times:
    assert trotter_steps > 0, f"{trotter_steps} must be > 0!"
    circuit = Circuit(n_qubits)
    for _i in range(trotter_steps):
        # Use the get_operators() generator to get the list of excitation operators
        for raw_pauli_string in qubit_ucc_operator.get_operators():
            # Convert each operator into a string and get the associated coefficient
            ((pauli_ops, coeff),) = raw_pauli_string.terms.items()  # Unpack the single-item dictionary
            pauli_string = " ".join(f"{pauli_op[1]}{pauli_op[0]}" for pauli_op in pauli_ops)
            # Build the circuit and add it on
            _circuit = expi_pauli(
                n_qubits, pauli_string, -1.0j * coeff * theta / trotter_steps
            )  # Divide imag. coeff by 1.0j
            circuit += _circuit
    return circuit


# Utility functions


def mp2_amplitude(excitation, orbital_energies, tei):
    r"""
    Calculate the MP2 guess amplitude for a single UCC circuit: 0.0 for a single excitation.
        for a double excitation (In SO basis): :math:`t_{ij}^{ab} = (g_{ijab} - g_{ijba}) / (e_i + e_j - e_a - e_b)`

    Args:
        excitation: Iterable of spin-orbitals representing a excitation. Must have either 2 or 4 elements exactly,
            representing a single or double excitation respectively.
        orbital_energies: eigenvalues of the Fock operator, i.e. orbital energies
        tei: Two-electron integrals in MO basis and second quantization notation

    Returns:
        MP2 guess amplitude (float)
    """
    # Checks validity of excitation argument
    assert len(excitation) % 2 == 0 and len(excitation) // 2 <= 2, f"{excitation} must have either 2 or 4 elements"
    # If single excitation, can just return 0.0 directly
    if len(excitation) == 2:
        return 0.0

    # Convert orbital indices to be in MO basis
    mo_orbitals = [orbital // 2 for orbital in excitation]
    # Numerator: g_ijab - g_ijba
    g_ijab = (
        tei[tuple(mo_orbitals)]  # Can index directly using the MO TEIs
        if (excitation[0] + excitation[3]) % 2 == 0 and (excitation[1] + excitation[2]) % 2 == 0
        else 0.0
    )
    g_ijba = (
        tei[tuple(mo_orbitals[:2] + mo_orbitals[2:][::-1])]  # Reverse last two terms
        if (excitation[0] + excitation[2]) % 2 == 0 and (excitation[1] + excitation[3]) % 2 == 0
        else 0.0
    )
    numerator = g_ijab - g_ijba
    # Denominator is directly from the orbital energies
    denominator = sum(orbital_energies[mo_orbitals[:2]]) - sum(orbital_energies[mo_orbitals[2:]])
    return numerator / denominator


def generate_excitations(order, excite_from, excite_to, conserve_spin=True):
    """
    Generate all possible excitations between a list of occupied and virtual orbitals

    Args:
        order: Order of excitations, i.e. 1 == single, 2 == double
        excite_from: Iterable of integers
        excite_to: Iterable of integers
        conserve_spin: ensure that the total electronic spin is conserved

    Return:
        List of lists, e.g. [[0, 1]]
    """
    # If order of excitation > either number of electrons/orbitals, return list of empty list
    if order > min(len(excite_from), len(excite_to)):
        return [[]]

    # Generate all possible excitations first
    from itertools import combinations  # pylint: disable=C0415

    all_excitations = [
        [*_from, *_to] for _from in combinations(excite_from, order) for _to in combinations(excite_to, order)
    ]
    # Filter out the excitations if conserve_spin set
    if conserve_spin:
        # Not sure if this filtering is exhaustive; might not remove some redundant excitations?
        all_excitations = [
            _ex
            for _ex in all_excitations
            if sum(_ex) % 2 == 0 and (sum(_i % 2 for _i in _ex[:order]) == sum(_i % 2 for _i in _ex[order:]))
        ]
    return all_excitations


def sort_excitations(excitations):
    """
    Sorts a list of excitations according to some common-sense and empirical rules (see below).
        The order of the excitations must be the same throughout.

    Sorting order:
    1. (For double excitations only) All paired excitations between the same MOs first
    2. Pair up excitations between the same MOs, e.g. (0, 2) and (1, 3)
    3. Then count upwards from smallest MO

    Args:
        excitations: List of iterables, each representing an excitation; e.g. [[1, 5], [0, 4]]

    Returns:
        List of excitations after sorting
    """
    # Check that all excitations are of the same order and <= 2
    order = len(excitations[0]) // 2
    if order > 2:
        raise NotImplementedError("Can only handle single and double excitations!")
    assert all(len(_ex) // 2 == order for _ex in excitations), "Cannot handle excitations of different orders!"

    # Define variables for the while loop
    copy_excitations = [list(_ex) for _ex in excitations]
    result = []
    prev = []

    # No idea how I came up with this, but it seems to work for double excitations
    def sorting_fn(x):
        # Default sorting is OK for single excitations
        return sum((order + 1 - _i) * abs(x[2 * _i + 1] // 2 - x[2 * _i] // 2) for _i in range(0, order))

    # Make a copy of the list of excitations, and use it populate a new list iteratively
    while copy_excitations:
        if not prev:
            # Take out all pair excitations first
            pair_excitations = [
                _ex
                for _ex in copy_excitations
                # Indices of the electrons/holes must be consecutive numbers
                if sum(abs(_ex[2 * _i + 1] // 2 - _ex[2 * _i] // 2) for _i in range(0, order)) == 0
            ]
            while pair_excitations:
                pair_excitations = sorted(pair_excitations)
                ex_to_remove = pair_excitations.pop(0)
                if ex_to_remove in copy_excitations:
                    # 'Move' the first pair excitation from copy_excitations to result
                    index = copy_excitations.index(ex_to_remove)
                    result.append(copy_excitations.pop(index))

            # No more pair excitations, only remaining excitations should have >=3 MOs involved
            # Sort the remaining excitations
            copy_excitations = sorted(copy_excitations, key=sorting_fn if order != 1 else None)
        else:
            # Check to see for excitations involving the same MOs as prev
            _from = prev[:order]
            _to = prev[order:]
            # Get all possible excitations involving the same MOs
            new_from = [_i + 1 if _i % 2 == 0 else _i - 1 for _i in _from]
            new_to = [_i + 1 if _i % 2 == 0 else _i - 1 for _i in _to]
            same_mo_ex = [sorted(list(_f) + list(_t)) for _f in (_from, new_from) for _t in (_to, new_to)]
            # Remove the excitations with the same MOs from copy_excitations
            while same_mo_ex:
                same_mo_ex = sorted(same_mo_ex)
                ex_to_remove = same_mo_ex.pop(0)
                if ex_to_remove in copy_excitations:
                    # 'Move' the first entry of same_mo_index from copy_excitations to result
                    index = copy_excitations.index(ex_to_remove)
                    result.append(copy_excitations.pop(index))
            prev = None
            continue
        if copy_excitations:
            # Remove the first entry from the sorted list of remaining excitations and add it to result
            prev = copy_excitations.pop(0)
            result.append(prev)
    return result


def ucc_ansatz(
    molecule,
    excitation_level=None,
    excitations=None,
    thetas=None,
    trotter_steps=1,
    ferm_qubit_map=None,
    include_hf=True,
    use_mp2_guess=True,
):
    """
    Convenience function for buildng a circuit corresponding to the UCC ansatz with multiple excitations for a given ``Molecule``.
    If no excitations are given, it defaults to returning the full UCCSD circuit ansatz.

    Args:
        molecule: The ``Molecule`` of interest.
        excitation_level: Include excitations up to how many electrons, i.e. ``"S"`` or ``"D"``.
            Ignored if ``excitations`` argument is given. Default: ``"D"``, i.e. double excitations
        excitations: List of excitations (e.g. ``[[0, 1, 2, 3], [0, 1, 4, 5]]``) used to build the
            UCC circuit. Overrides the ``excitation_level`` argument
        thetas: Parameters for the excitations. Default value depends on the ``use_mp2_guess`` argument.
        trotter_steps: number of Trotter steps; i.e. number of times the UCC ansatz is applied with
            ``theta`` = ``theta / trotter_steps``. Default: 1
        ferm_qubit_map: fermion-to-qubit transformation. Default: Jordan-Wigner (``"jw"``)
        include_hf: Whether or not to start the circuit with a Hartree-Fock circuit. Default: ``True``
        use_mp2_guess: Whether to use MP2 amplitudes or a numpy zero array as the initial guess parameter. Default: ``True``;
            use the MP2 amplitudes as the default guess parameters

    Returns:
        Qibo ``Circuit``: Circuit corresponding to an UCC ansatz
    """
    # Get the number of electrons and spin-orbitals from the molecule argument
    n_elec = molecule.nelec if molecule.n_active_e is None else molecule.n_active_e
    n_orbs = molecule.nso if molecule.n_active_orbs is None else molecule.n_active_orbs

    # Define the excitation level to be used if no excitations given
    if excitations is None:
        excitation_levels = ("S", "D", "T", "Q")
        if excitation_level is None:
            excitation_level = "D"
        else:
            # Check validity of input
            assert (
                len(excitation_level) == 1 and excitation_level.upper() in excitation_levels
            ), "Unknown input for excitation_level"
            # Note: Probably don't be too ambitious and try to do 'T'/'Q' at the moment...
            if excitation_level.upper() in ("T", "Q"):
                raise NotImplementedError("Cannot handle triple and quadruple excitations!")
        # Get the (largest) order of excitation to use
        excitation_order = excitation_levels.index(excitation_level.upper()) + 1

        # Generate and sort all the possible excitations
        excitations = []
        for order in range(excitation_order, 0, -1):  # Reversed to get higher excitations first
            excitations += sort_excitations(generate_excitations(order, range(0, n_elec), range(n_elec, n_orbs)))
    else:
        # Some checks to ensure the given excitations are valid
        assert all(len(_ex) % 2 == 0 for _ex in excitations), "Excitation with an odd number of elements found!"

    # Check if thetas argument given, define to be all zeros if not
    # Number of circuit parameters: S->2, D->8, (T/Q->32/128; Not sure?)
    n_parameters = 2 * len([_ex for _ex in excitations if len(_ex) == 2])  # Singles
    n_parameters += 8 * len([_ex for _ex in excitations if len(_ex) == 4])  # Doubles
    if thetas is None:
        if use_mp2_guess:
            thetas = np.array([mp2_amplitude(excitation, molecule.eps, molecule.tei) for excitation in excitations])
        else:
            thetas = np.zeros(n_parameters)
    else:
        # Check that number of circuit variables (i.e. thetas) matches the number of circuit parameters
        assert len(thetas) == n_parameters, "Number of input parameters doesn't match the number of circuit parameters!"

    # Build the circuit
    if include_hf:
        circuit = hf_circuit(n_orbs, n_elec, ferm_qubit_map=ferm_qubit_map)
    else:
        circuit = Circuit(n_orbs)
    for excitation, theta in zip(excitations, thetas):
        circuit += ucc_circuit(n_orbs, excitation, theta, trotter_steps=trotter_steps, ferm_qubit_map=ferm_qubit_map)
    return circuit



"""
Circuit representing a Hartree-Fock reference wave function
"""

import numpy as np
from qibo import gates, models


# Helper functions for the Brayvi-Kitaev mapping
def bk_matrix_power2(n: int):
    """Build the Brayvi-Kitaev matrix of dimension n (= power of 2) recursively

    Args:
        n: Size of BK matrix
    """
    assert n > 0, "Dimension of BK matrix must be at least 1"
    # Base case
    if n == 1:
        return np.ones((1, 1), dtype=np.int8)
    # Recursive definition
    smaller_bk_matrix = bk_matrix_power2(n - 1)
    top_right = np.zeros((2 ** (n - 2), 2 ** (n - 2)), dtype=np.int8)
    top_half = np.concatenate((smaller_bk_matrix, top_right), axis=1)

    bottom_left = np.concatenate(
        (np.zeros(((2 ** (n - 2)) - 1, 2 ** (n - 2)), dtype=np.int8), np.ones((1, 2 ** (n - 2)), dtype=np.int8)), axis=0
    )
    bottom_half = np.concatenate((bottom_left, smaller_bk_matrix), axis=1)
    # Combine top and bottom half
    return np.concatenate((top_half, bottom_half), axis=0)


def bk_matrix(n: int):
    """Exact Brayvi-Kitaev matrix of size n, obtained by slicing a larger BK matrix
        with dimension 2**m > n

    TODO: Update the occupation number vector using the update, parity, and flip set instead?
        Not sure if necessary; i.e. size of BK matrix probably not comparable to the memory needed
        for a classical simulation?

    Args:
        n: Size of BK matrix
    """
    assert n > 0, "Dimension of the Brayvi-Kitaev matrix must be at least 1"
    # Build bk_matrix_power2(m), where 2**m > n
    min_bk_size = int(np.ceil(np.log2(n))) + 1
    min_bk_matrix = bk_matrix_power2(min_bk_size)
    # Then use array slicing to get the actual BK matrix
    return min_bk_matrix[:n, :n]


# Main function
def hf_circuit(n_qubits, n_electrons, ferm_qubit_map=None):
    """Circuit to prepare a Hartree-Fock state

    Args:
        n_qubits: Number of qubits in the quantum circuit
        n_electrons: Number of electrons in the molecular system
        ferm_qubit_map: Fermion to qubit map. Must be either Jordan-Wigner (``jw``) or Brayvi-Kitaev (``bk``). Default value is ``jw``.

    Returns:
        Qibo ``Circuit``: Circuit initialized in a HF reference state
    """
    # Which fermion-to-qubit map to use
    if ferm_qubit_map is None:
        ferm_qubit_map = "jw"
    if ferm_qubit_map not in ("jw", "bk"):
        raise KeyError("Fermon-to-qubit mapping must be either 'jw' or 'bk'")

    # Occupation number of SOs
    occ_n = np.concatenate((np.ones(n_electrons), np.zeros(n_qubits - n_electrons)))
    if ferm_qubit_map == "jw":
        mapped_occ_n = occ_n
    elif ferm_qubit_map == "bk":
        mapped_occ_n = (bk_matrix(n_qubits) @ occ_n) % 2
    # Convert to ints
    mapped_occ_n = mapped_occ_n.astype(int)
    # Build the circuit using the mapped vector
    circuit = models.Circuit(n_qubits)
    circuit.add(gates.X(int(_i)) for _i, _mo in enumerate(mapped_occ_n) if _mo == 1)
    return circuit
